package uk.ac.ukc.cs.kmf.toolgen.vsml.repository.implementation;

import java.io.File;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.util.Collection;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.IdentityHashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.Vector;

import javax.swing.tree.DefaultMutableTreeNode;

import uk.ac.ukc.cs.kmf.toolgen.vsml.IvsmlElement;
import uk.ac.ukc.cs.kmf.toolgen.vsml.IvsmlLifecycle;
import uk.ac.ukc.cs.kmf.toolgen.vsml.repository.IHUTNVisitor;
import uk.ac.ukc.cs.kmf.toolgen.vsml.repository.IJTreeVisitor;
import uk.ac.ukc.cs.kmf.toolgen.vsml.repository.IXMIVisitor;
import uk.ac.ukc.cs.kmf.toolgen.vsml.repository.IvsmlRepository;
import uk.ac.ukc.cs.kmf.util.FileLog;
import uk.ac.ukc.cs.kmf.util.ILog;
import uk.ac.ukc.cs.kmf.xmi.AdapterFactoryRegister;
import uk.ac.ukc.cs.kmf.xmi.IXMIWriter;
import uk.ac.ukc.cs.kmf.xmi.XMIFile;
import uk.ac.ukc.cs.kmf.xmi.XMIObject;
import uk.ac.ukc.cs.kmf.xmi.XMIReader;
import uk.ac.ukc.cs.kmf.xmi.XMIWriter;

public class vsmlRepository
  implements IvsmlRepository
{
  //--- CONSTRUCTOR ---
  public vsmlRepository() {
    setLog(new FileLog("Repository.log"));
    init();
  }
  public vsmlRepository(ILog log) {
    setLog(log);
    init();
  }

   // Init the attributes
   protected void init() {
     log.reportMessage("Init vsml repository");
     // Add lifecycles
     log.reportMessage("Adding lifecycle and initializing population for 'uk.ac.ukc.cs.kmf.toolgen.vsml.AS.Class_'");
     lifecycles.put("uk.ac.ukc.cs.kmf.toolgen.vsml.AS.Class_", (IvsmlLifecycle)new uk.ac.ukc.cs.kmf.toolgen.vsml.AS.implementation.Class_Lifecycle(this));
     population.put("uk.ac.ukc.cs.kmf.toolgen.vsml.AS.Class_", new Vector());
     log.reportMessage("Adding lifecycle and initializing population for 'uk.ac.ukc.cs.kmf.toolgen.vsml.AS.AssociationEnd'");
     lifecycles.put("uk.ac.ukc.cs.kmf.toolgen.vsml.AS.AssociationEnd", (IvsmlLifecycle)new uk.ac.ukc.cs.kmf.toolgen.vsml.AS.implementation.AssociationEndLifecycle(this));
     population.put("uk.ac.ukc.cs.kmf.toolgen.vsml.AS.AssociationEnd", new Vector());
     log.reportMessage("Adding lifecycle and initializing population for 'uk.ac.ukc.cs.kmf.toolgen.vsml.AS.Package_'");
     lifecycles.put("uk.ac.ukc.cs.kmf.toolgen.vsml.AS.Package_", (IvsmlLifecycle)new uk.ac.ukc.cs.kmf.toolgen.vsml.AS.implementation.Package_Lifecycle(this));
     population.put("uk.ac.ukc.cs.kmf.toolgen.vsml.AS.Package_", new Vector());
     log.reportMessage("Adding lifecycle and initializing population for 'uk.ac.ukc.cs.kmf.toolgen.vsml.AS.Association'");
     lifecycles.put("uk.ac.ukc.cs.kmf.toolgen.vsml.AS.Association", (IvsmlLifecycle)new uk.ac.ukc.cs.kmf.toolgen.vsml.AS.implementation.AssociationLifecycle(this));
     population.put("uk.ac.ukc.cs.kmf.toolgen.vsml.AS.Association", new Vector());

     elemId = 0;
   }

   //--- Build an element ---
   public IvsmlElement buildElement(String fullClassName) {
     IvsmlElement elem = null;
     try {
       // Deal with enumerators
       String ownerName = new String(fullClassName);
       int pos = ownerName.lastIndexOf('.');
       if (pos != -1) ownerName = ownerName.substring(0, pos);
       if (ownerName.endsWith("Enum") || ownerName.endsWith("Kind")) {
          String enumName = fullClassName.substring(pos+1, fullClassName.length());
          java.lang.Class cls = java.lang.Class.forName(ownerName);
          elem = (IvsmlElement)cls.getField(enumName).get(null);
       } else {
          IvsmlLifecycle lifecycle = (IvsmlLifecycle)lifecycles.get(fullClassName);
          elem = lifecycle.build();
       }
     } catch (Exception e) {
        log.reportMessage("Error: Missing lifecycle '"+fullClassName+"'");
     }
     return elem;
   }

   //--- Add an element ---
   public void addElement(String fullClassName, IvsmlElement elem) {
     try {
       ((List)population.get(fullClassName)).add(elem);
     } catch (Exception e) {;
       log.reportMessage("Error: Missing population for '"+fullClassName+"'");
     }
   }

   //--- Remove an element ---
   public void removeElement(String fullClassName, IvsmlElement elem) {
     try {
       ((List)population.get(fullClassName)).remove(elem);
     } catch (Exception e) {;
       log.reportMessage("Error: Missing population for '"+fullClassName+"'");
     }
   }

   //--- Get all elements of a type ---
   public List getElements(String fullClassName) {
     List objs = null;
     try {
       objs = (List)population.get(fullClassName);
     } catch (Exception e) {;
            log.reportMessage("Error: Missing population for '"+fullClassName+"' class");
     }
     return objs;
   }

   //--- Check if the first class name is an instance of the second class name ---
   protected boolean isInstanceOf(String elemName, String fullClassName) {
     //--- Make interfaces ---
     int pos1 = elemName.lastIndexOf('.');
     String iElemName = elemName;
     if (pos1 != -1 && elemName.charAt(pos1+1) != 'I')
       iElemName = elemName.substring(0, pos1+1)+"I"+elemName.substring(pos1+1, elemName.length()); 
     int pos2 = fullClassName.lastIndexOf('.');
     String iFullClassName = fullClassName;
     if (pos2 != -1 && fullClassName.charAt(pos2+1) != 'I')
       iFullClassName = fullClassName.substring(0, pos2+1)+"I"+fullClassName.substring(pos2+1, fullClassName.length());
     //--- Check if fullClassName is a superinterface of elemName ---
     boolean res = false;
     try {
       res = Class.forName(iFullClassName).isAssignableFrom(Class.forName(iElemName));
     } catch (Exception e) {
     }
     return res;
   }
   //--- Get all instances of a type ---
   public List getInstances(String fullClassName) {
     List objs = new Vector();
     Set elemNames = population.keySet();
     Iterator nameIt = elemNames.iterator();
     while (nameIt.hasNext()) {
       String elemName = (String)nameIt.next();
       if (isInstanceOf(elemName, fullClassName)) {
         List elements = (List)population.get(elemName);
         Iterator elemIt = elements.iterator();
         while (elemIt.hasNext()) {
           Object elem = elemIt.next();
           objs.add(elem);
         }
       }
     }
     return objs;
   }

   //--- Get a specific lifecycle ---
   public IvsmlLifecycle getLifecycle(String fullClassName) {
     IvsmlLifecycle lifecycle = null;
     try {
        lifecycle = (IvsmlLifecycle)lifecycles.get(fullClassName);
     } catch (Exception e) {
            log.reportMessage("Error: Missing factory for '"+fullClassName+"' class");
     }
     return lifecycle;
   }

   //--- Add a lifecycle ---
   public void addLifecycle(String fullClassName, IvsmlLifecycle lifecycle) {
     if (!lifecycles.containsKey(fullClassName)) {
       log.reportMessage("Warning: Lifecycle for '"+fullClassName+"' was replaced");
     }
     lifecycles.put(fullClassName, lifecycle);
   }

   //--- HUTN ---
   public String toHUTN() {
     //--- Creat the HUTN visitor ---
     IHUTNVisitor hutnVisitor = new HUTNVisitor();
     //--- Add every lifecycle ---
     String buffer = new String();
     buffer += "vsmlRepository {\n";
     buffer += "Lifecycles\n";
     Set lifecycleNames = lifecycles.keySet();
     Iterator lifecycleIt = lifecycleNames.iterator();
     while (lifecycleIt.hasNext()) {
       String lifecycleName = (String)lifecycleIt.next();
       buffer += "  "+lifecycleName+"\n";
     }
     buffer += "\n";
     //--- Add every element ---
     buffer += "Elements\n";
     Set elemNames = population.keySet();
     Iterator elemIt = elemNames.iterator();
     while (elemIt.hasNext()) {
       String elemName = (String)elemIt.next();
       buffer += "Instances of "+elemName+"\n";
       List elements = (List)population.get(elemName);
       Iterator elementIt = elements.iterator();
       while (elementIt.hasNext()) {
         IvsmlElement element = (IvsmlElement)elementIt.next();
         buffer += element.accept(hutnVisitor, new Object());
       }
     }
     buffer += "}\n";
     log.reportMessage(buffer);
     return buffer;
   }

   public void saveHUTN(String fileName) {
     //--- Open the HUTN file ---
     if (!fileName.toLowerCase().endsWith(".hutn")) fileName += ".hutn";
     PrintWriter out;
     try {
       out = new PrintWriter(new FileWriter(new File(fileName)), true);
     } catch (Exception e) {
       log.reportMessage(""+e);
       return;
     }
     //--- Creat the HUTN visitor ---
     IHUTNVisitor hutnVisitor = new HUTNVisitor();
     //--- Add every lifecycle ---
     out.println("vsmlRepository {");
     out.println("Lifecycles");
     Set lifecycleNames = lifecycles.keySet();
     Iterator lifecycleIt = lifecycleNames.iterator();
     while (lifecycleIt.hasNext()) {
       String lifecycleName = (String)lifecycleIt.next();
       out.println("  "+lifecycleName);
     }
     out.println();
     //--- Add every element ---
     out.println("Elements");
     Set elemNames = population.keySet();
     Iterator elemIt = elemNames.iterator();
     while (elemIt.hasNext()) {
       String elemName = (String)elemIt.next();
       out.println("Instances of "+elemName);
       List elements = (List)population.get(elemName);
       Iterator elementIt = elements.iterator();
       while (elementIt.hasNext()) {
         IvsmlElement element = (IvsmlElement)elementIt.next();
         out.println(element.accept(hutnVisitor, new Object()));
       }
     }
     out.println("}");
     out.close();
   }
   public IvsmlRepository loadHUTN(String fileName) {
     try {
       //--- Create a repository ---
       vsmlRepository rep = new vsmlRepository();
       return rep;
     } catch (Exception e) {
       log.reportMessage(""+e);
       return null;
     }
   }

   //--- toJTree ---
   public DefaultMutableTreeNode toJTree() {
     //--- Creat the JTree visitor ---
     IJTreeVisitor jtreeVisitor = new JTreeVisitor();
     //--- Init root ---
     DefaultMutableTreeNode root = new DefaultMutableTreeNode("Repository", true);
     DefaultMutableTreeNode lifecyclesNode = new DefaultMutableTreeNode("Lifecycles", true);
     DefaultMutableTreeNode elementsNode = new DefaultMutableTreeNode("Elements", true);
     root.add(lifecyclesNode);
     root.add(elementsNode);
     //--- Add every lifecycle ---
     Set lifecycleNames = lifecycles.keySet();
     Iterator lifecycleIt = lifecycleNames.iterator();
     while (lifecycleIt.hasNext()) {
       String lifecycleName = (String)lifecycleIt.next();
       IvsmlLifecycle lifecycle = (IvsmlLifecycle)lifecycles.get(lifecycleName);
       DefaultMutableTreeNode lifecycleNode = (DefaultMutableTreeNode)lifecycle.accept(jtreeVisitor, new Object());
       lifecyclesNode.add(lifecycleNode);
     }
     //--- Add every element ---
     Set elemNames = population.keySet();
     Iterator elemIt = elemNames.iterator();
     while (elemIt.hasNext()) {
       String elemName = (String)elemIt.next();
       //--- Add a node for each class ---
       List elements = (List)population.get(elemName);
       if (elements.size() != 0) {
         //Find or create node for this element class
         DefaultMutableTreeNode currNode = elementsNode;
         StringTokenizer tokens = new StringTokenizer(elemName,".");
         while (tokens.hasMoreTokens()) {
           String token = tokens.nextToken();
           DefaultMutableTreeNode node = null;
           Enumeration en = currNode.children();
           while (en.hasMoreElements()) {
             Object o = en.nextElement();
             if (o.toString().equals(token)) {
               node = (DefaultMutableTreeNode)o;
             }
           }
           if (node == null) {
             node  = new DefaultMutableTreeNode(token, true);
             currNode.add(node);
           }
           currNode = node;
         }
         DefaultMutableTreeNode elemNode = currNode;
         Iterator elementIt = elements.iterator();
         while (elementIt.hasNext()) {
           IvsmlElement element = (IvsmlElement)elementIt.next();
           //--- Add a node for each instance ---
           DefaultMutableTreeNode instanceNode = (DefaultMutableTreeNode)element.accept(jtreeVisitor, new Object());
           elemNode.add(instanceNode);
         }
         //elementsNode.add(elemNode);
       }
     }
     return root;
   }

   //--- ParseAll/EvaluateAll ---
   public void inv_ParseAll() {
     log.reportMessage("");
     //--- Parse every invariant ---
     log.reportMessage("Parsing invariants from repository ...");
     Set elemNames = population.keySet();
     Iterator elemIt = elemNames.iterator();
     while (elemIt.hasNext()) {
       String elemName = (String)elemIt.next();
       List elements = (List)population.get(elemName);
       Iterator elementIt = elements.iterator();
       while (elementIt.hasNext()) {
         IvsmlElement element = (IvsmlElement)elementIt.next();
         element.inv_ParseAll(log);
       }
     }
   }
   public void inv_EvaluateAll() {
     log.reportMessage("");
     //--- Evaluate every invariant ---
     log.reportMessage("Evaluating invariants from repository ...");
     Set elemNames = population.keySet();
     Iterator elemIt = elemNames.iterator();
     while (elemIt.hasNext()) {
       String elemName = (String)elemIt.next();
       List elements = (List)population.get(elemName);
       Iterator elementIt = elements.iterator();
       while (elementIt.hasNext()) {
         IvsmlElement element = (IvsmlElement)elementIt.next();
         element.inv_EvaluateAll(log);
       }
     }
   }

   //--- XMI ---
   public void saveXMI(String fileName) {
     //--- Create an XMI file ---
     if (!fileName.toLowerCase().endsWith(".xmi")) fileName += ".xmi";
     XMIFile xmiFile = new XMIFile(fileName);
     //--- Create an XMI vistor ---
     IXMIVisitor xmiVisitor = new XMIVisitor();
     //--- Create an empty (reference, xmiObject) table  ---
     Map mapRefToXMI = new IdentityHashMap();
     //--- Create the top objects container ---
     Collection topObjs = xmiFile.getTopObjects();
     //--- For each lifecycle ---
     Set lifecycleNames = lifecycles.keySet();
     Iterator lifecycleIt = lifecycleNames.iterator();
     while (lifecycleIt.hasNext()) {
       String elemName = (String)lifecycleIt.next();
       IvsmlLifecycle lifecycle = (IvsmlLifecycle)lifecycles.get(elemName);
       //--- Create the lifecycle ---
       XMIObject xmiLifecycle = (XMIObject)lifecycle.accept(xmiVisitor, mapRefToXMI);
       //--- Add lifecycle to top objects ---
       //--- Compute all the instances created by the lifecycle ---
       List elements = (List)population.get(elemName);
       Iterator elementIt = elements.iterator();
       while (elementIt.hasNext()) {
         IvsmlElement element = (IvsmlElement)elementIt.next();
         //--- Create the XMI element ---
         XMIObject xmiElement = (XMIObject)element.accept(xmiVisitor, mapRefToXMI);
         //--- Add the element to top objects ---
         topObjs.add(xmiElement);
       }
     }
     try {
       //--- Register the AdapterFactory into XMI framework ---
       AdapterFactoryRegister.registerAdapterFactory(new vsmlAdapterFactory(null));
       //--- Save the elements ---
       IXMIWriter writer = new XMIWriter();
       writer.write(xmiFile, fileName, log);
     } catch (Exception e) {
            log.reportMessage(""+e);
     }
   }
   public IvsmlRepository loadXMI(String fileName) {
     try {
       //--- Create a repository ---
       vsmlRepository rep = new vsmlRepository();
       rep.setLog(log);
       //--- Register the AdapterFactory into XMI framework ---
       AdapterFactoryRegister.registerAdapterFactory(new vsmlAdapterFactory(rep));
       //--- Build the elements ---
       XMIFile xmiFile = (new XMIReader()).read(fileName, log);
       return rep;
     } catch (Exception e) {
       log.reportMessage(""+e);
       return null;
     }
   }

   //--- LOG ---
   public ILog getLog() {
     return log;
   }
   public void setLog(ILog log) {
     this.log = log;
   }

   //--- Ids ---
   protected static int elemId = 0;
   public static void resetId() {
     elemId = 0;
   }
   public String newId() {
     return ""+ ++elemId;
   }

   //--- LOG ---
   protected ILog log;
   //--- Lifecycles ---
   protected Hashtable lifecycles = new Hashtable();
   //--- Population ---
   protected Hashtable population = new Hashtable();

}
